<!DOCTYPE html>
<html>
<head lang="en">
    <meta charset="UTF-8">
    <title>mitmproxy</title>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="stylesheet" href="/static/vendor.css"/>
    <link rel="stylesheet" href="/static/app.css"/>
    <link rel="icon" href="/static/images/favicon.ico" type="image/x-icon"/>
    <script src="/conf.js"></script>
    <script src="/static/app.js"></script>
</head>
<body>
<div id="mitmproxy"></div>
</body>
</html>
